/*
 * Copyright 2013 Malcolm Lett
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sourceforge.tedhi;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Locale.Category;
import java.util.Map;

/**
 * Parses date patterns that can be passed by {@link java.text.SimpleDateFormat}.
 * The internal workings are loosely based on private {@code SimpleDateFormat.compile()}
 * method.
 * 
 * <p> Instances are immutable and thread-safe.
 */
class DatePattern {
	private final static String TAG_CHARS = DateTimeFieldType.codeString();
	
	private String pattern;
	private List<PatternPart> parsedPattern;
	private List<PatternPart> parsedPatternWithContractedLiterals; // runs of consecutive literals combined
	private List<Tag> tags;

	/**
	 * Gets the date/time pattern with the default formatting style for the default locale.
	 * @return a date pattern.
	 */
	public static DatePattern getInstance() {
		return forDateFormat(DateFormat.getInstance());
	}
	
	/**
	 * Gets the date pattern with the default formatting style for the default locale.
	 * @return a date pattern.
	 */
	public static DatePattern getDateInstance() {
		return forDateFormat(DateFormat.getDateInstance());
	}
	
	/**
	 * Gets the date pattern with the given formatting style for the default locale.
	 * @param dateStyle the given formatting style. For example, SHORT for "M/d/yy" in the US locale.
	 * @return a date pattern.
	 */
	public static DatePattern getDateInstance(int dateStyle) {
		return forDateFormat(DateFormat.getDateInstance(dateStyle, Locale.getDefault(Category.FORMAT)));
	}

	/**
	 * Gets the date pattern with the given formatting style for the given locale.
	 * @param dateStyle the given formatting style. For example, SHORT for "M/d/yy" in the US locale.
	 * @param locale the given locale
	 * @return a date pattern.
	 */
	public static DatePattern getDateInstance(int dateStyle, Locale locale) {
		return forDateFormat(DateFormat.getDateInstance(dateStyle, locale));
	}
	
	/**
	 * Gets the time pattern with the default formatting style for the default locale.
	 * @return a time pattern.
	 */
	public static DatePattern getTimeInstance() {
		return forDateFormat(DateFormat.getTimeInstance());
	}
	
	/**
	 * Gets the time pattern with the given formatting style for the default locale.
	 * @param timeStyle the given formatting style. For example, SHORT for "h:mm a" in the US locale.
	 * @return a time pattern.
	 */
	public static DatePattern getTimeInstance(int timeStyle) {
		return forDateFormat(DateFormat.getTimeInstance(timeStyle, Locale.getDefault(Category.FORMAT)));
	}

	/**
	 * Gets the time pattern with the given formatting style for the given locale.
	 * @param timeStyle the given formatting style. For example, SHORT for "h:mm a" in the US locale.
	 * @param locale the given locale
	 * @return a time pattern.
	 */
	public static DatePattern getTimeInstance(int timeStyle, Locale locale) {
		return forDateFormat(DateFormat.getTimeInstance(timeStyle, locale));
	}
	
	/**
	 * Gets the date pattern with the given formatting style for the default locale.
	 * @param dateStyle the given formatting style. For example, SHORT for "M/d/yy" in the US locale.
	 * @param timeStyle the given formatting style. For example, SHORT for "h:mm a" in the US locale.
	 * @return a date/time pattern.
	 */
	public static DatePattern getDateTimeInstance(int dateStyle, int timeStyle) {
		return forDateFormat(DateFormat.getDateTimeInstance(dateStyle, timeStyle, Locale.getDefault(Category.FORMAT)));
	}

	/**
	 * Gets the date pattern with the given formatting style for the given locale.
	 * @param dateStyle the given formatting style. For example, SHORT for "M/d/yy" in the US locale.
	 * @param timeStyle the given formatting style. For example, SHORT for "h:mm a" in the US locale.
	 * @param locale the given locale
	 * @return a date/time pattern.
	 */
	public static DatePattern getDateTimeInstance(int dateStyle, int timeStyle, Locale locale) {
		return forDateFormat(DateFormat.getDateTimeInstance(dateStyle, timeStyle, locale));
	}
	
	/**
	 * Determines the date pattern given the supplied date format. 
	 * @param dateFormat
	 * @return a date/time pattern
	 */
	public static DatePattern forDateFormat(DateFormat dateFormat) {
		if (dateFormat instanceof SimpleDateFormat) {
			return new DatePattern(((SimpleDateFormat) dateFormat).toPattern());
		}
//		else {
//			DatePattern pattern = forDateFormatViaAttributes(dateFormat);
//			if (pattern != null) {
//				return pattern;
//			}
//		}
		
		// failure
		throw new IllegalArgumentException("Unable to determine pattern from "+dateFormat.getClass().getName());
	}
	
// TODO: complete implementation as full-back mechanism for when date format providers
// return something other than SimpleDateFormat
// TODO: probably return null if formatToCharacterIterator() returns no attributes
// and let caller raise the exception.
//	private static DatePattern forDateFormatViaAttributes(DateFormat dateFormat) {
//		AttributedCharacterIterator itr = dateFormat.formatToCharacterIterator(new Date());
//		Run currentRun = Run.of(itr);
//		List<PatternPart> parts = new ArrayList<PatternPart>();
//		while (itr.current() != AttributedCharacterIterator.DONE) {
//			// complete previous run
//			if (!currentRun.sameAsFor(itr)) {
//				// TODO: add to result
//				
//				// start new run
//				currentRun = Run.of(itr);
//			}
//			
//			
//			
//			// TODO: get attribute by scanning through list of associated attributes
//			
//			char ch = itr.current();
//			System.out.println("  '"+ch+"': @"+itr.getIndex()+", "+itr.getAttributes()+" (run: "+itr.getRunStart()+" to "+itr.getRunLimit()+")");
//			itr.next();
//		}
//		char ch = itr.current(); 
//		System.out.println("  '"+ch+"': @"+itr.getIndex()+", "+itr.getAttributes()+" (run: "+itr.getRunStart()+" to "+itr.getRunLimit()+")");
//	}
	
//	private static class Run {
//		private int start;
//		private int limit;
//		private DatePatternLetter letter;
//		
//		public static Run of(AttributedCharacterIterator itr) {
//			Run run = new Run();
//			run.start = itr.getRunStart();
//			run.limit = itr.getRunLimit();
//			
//			for (Attribute attr: itr.getAllAttributeKeys()) {
//				if (attr instanceof DateFormat.Field) {
//					DateFormat.Field field = (DateFormat.Field) attr;
//					int calendarField = field.getCalendarField();
//					if (calendarField >= 0) {
//						run.letter = DatePatternLetter.valueOfCalendarField(calendarField);
//						// TODO: for hour fields, have to check whether it's returned the current hour or +/- 1.
//						// TODO: or use the "hour 1" vs. "hour 0" name of the field  ;-\
//					}
//				}
//			}
//			return run;
//		}
//		
//		public boolean sameAsFor(AttributedCharacterIterator itr) {
//			return (start == itr.getRunStart()) && (limit == itr.getRunLimit());
//			// TODO: also check field just to make sure we pick up on weird cases
//		}
//	}
	
	/**
	 * Creates new instance given the pattern.
	 * @param pattern
	 */
	public DatePattern(String pattern) {
		this(parse(pattern));
	}

	/**
	 * Creates a new instance to wrap the pre-computed parts.
	 * @param parts
	 */
	public DatePattern(List<PatternPart> parts) {
		this.parsedPattern = Collections.unmodifiableList(new ArrayList<PatternPart>(parts));
		this.parsedPatternWithContractedLiterals = Collections.unmodifiableList(contractLiteralsOf(parts));

		List<Tag> tags = new ArrayList<Tag>();
		for (PatternPart part: parsedPattern) {
			if (part instanceof Tag) {
				tags.add((Tag) part);
			}
		}
		this.tags = Collections.unmodifiableList(tags);
		this.pattern = patternStringOf(parsedPattern);
	}
	
	/**
	 * Gets the original pattern.
	 * Calls {@link #toPatternString()}.
	 */
	@Override
	public String toString() {
		return toPatternString();
	}

	/**
	 * Generates a string symbolising the parsed data in detail.
	 * @return strings of the form such as {@code "TAG(d)LITERAL(/)TAG(MM)LITERAL(/)TAG(yyyy)"}
	 */
	public String toSymbolicString() {
		StringBuilder buf = new StringBuilder();
		for (PatternPart part: parsedPattern) {
			buf.append(part.toSymbolicString());
		}
		return buf.toString();
	}

	/**
	 * Gets the original pattern.
	 * @return strings of the form such as {@code "d/MM/yyyy"}
	 */
	public String toPatternString() {
		return pattern;
	}
	
	/**
	 * Generates the original pattern string.
	 * @return strings of the form such as {@code "d/MM/yyyy"}
	 */
	protected static String patternStringOf(List<PatternPart> parts) {
		StringBuilder buf = new StringBuilder();
		for (PatternPart part: parts) {
			buf.append(part.toPatternString());
		}
		return buf.toString();
	}
	
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((parsedPattern == null) ? 0 : parsedPattern.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (!(obj instanceof DatePattern)) {
			return false;
		}
		DatePattern other = (DatePattern) obj;
		if (parsedPattern == null) {
			if (other.parsedPattern != null) {
				return false;
			}
		} else if (!parsedPattern.equals(other.parsedPattern)) {
			return false;
		}
		return true;
	}

	public List<PatternPart> parts() {
		return parsedPattern;
	}

	public List<PatternPart> parts(boolean contractRunsOfLiterals) {
		if (contractRunsOfLiterals) {
			return parsedPatternWithContractedLiterals;
		}
		else {
			return parsedPattern;
		}
	}

	public List<Tag> tags() {
		return tags;
	}
	
	/**
	 * Indicates whether or not this contains a year.
	 */
	public boolean hasYear() {
		return firstYearTag() != null;
	}

	/**
	 * Indicates whether or not this contains a month.
	 */
	public boolean hasMonth() {
		return firstMonthTag() != null;
	}

	/**
	 * Indicates whether or not this contains a day.
	 */
	public boolean hasDay() {
		return firstDayTag() != null;
	}
	
	/**
	 * Gets the first literal, if one is present
	 * @return the literal or null if not present
	 */
	public Literal firstLiteral() {
		for (PatternPart part: parsedPattern) {
			if (part instanceof Literal) {
				return (Literal) part;
			}
		}
		return null;
	}
	
	/**
	 * Gets the first year tag, if one is present.
	 * @return the tag or null if not present
	 */
	public Tag firstYearTag() {
		for (Tag tag: tags()) {
			if (tag.type() == DateTimeFieldType.YEAR) {
				return tag;
			}
		}
		return null;
	}

	/**
	 * Gets the first month tag, if one is present.
	 * @return the tag or null if not present
	 */
	public Tag firstMonthTag() {
		for (Tag tag: tags()) {
			if (tag.type() == DateTimeFieldType.MONTH_IN_YEAR) {
				return tag;
			}
		}
		return null;
	}

	/**
	 * Gets the first day tag, if one is present.
	 * @return the tag or null if not present
	 */
	public Tag firstDayTag() {
		for (Tag tag: tags()) {
			if (tag.type() == DateTimeFieldType.DAY_IN_MONTH) {
				return tag;
			}
		}
		return null;
	}
	
	/**
	 * Derives a new date pattern by ordering the tags according to the
	 * order of {@code fieldOrder}, and by keeping associated literals with their tags.
	 * It is illegal for the current pattern to contain a field not specified
	 * in {@code fieldOrder}, however it is acceptable for {@code fieldOrder} to
	 * reference fields that are not present in this date pattern.
	 * 
	 * <p> Note: The implementation of this is currently very simplistic
	 * because it can afford to be given the way it's used.
	 * At present only {@code FlexibleDateFormat.ParserItemFactory.localizedOrderOf()}
	 * makes calls to this, which only ever passes patterns that have been hard-coded
	 * and use a simple consistent separator.
	 * Furthermore, it assumes this is a pattern with alternating tag/literals, that
	 * starts and ends with tags.
	 * @throws IllegalArgumentException if this pattern contains any tags that aren't
	 * 	       explicitly ordered by {@code fieldOrder}
	 * @throws UnsupportedOperationException if this pattern is too complex for this method's assumptions
	 */
	public DatePattern withFieldsOrderedAs(List<DateTimeFieldType> fieldOrder) {
		// get single consistent separator
		Literal separator = null;
		for (PatternPart part: parts(true)) {
			if (part instanceof Literal) {
				if (separator == null) {
					separator = (Literal) part;
				}
				else if (!separator.equals(part)) {
					// indicates a bug: the assumptions stated above are no longer valid
					throw new UnsupportedOperationException("not able to handle different separators, found ("+separator+", "+part+") in \""+toPatternString()+"\"");
				}
			}
		}
		
		// prepare for ordering
		Map<DateTimeFieldType, Tag> tagsToAddByType = new HashMap<DateTimeFieldType, Tag>();
		for (Tag tag: tags()) {
			if (!fieldOrder.contains(tag.type())) {
				throw new IllegalArgumentException("required order of "+tag.type()+" field not specified");
			}
			if (tagsToAddByType.containsKey(tag.type())) {
				// indicates a bug: the assumptions stated above are no longer valid
				throw new UnsupportedOperationException("not able to handle duplicated fields, found two or more "+tag.type()+" fields in \""+toPatternString()+"\"");
			}
			tagsToAddByType.put(tag.type(), tag);
		}

		// construct new ordered pattern
		List<PatternPart> newParts = new ArrayList<PatternPart>();
		for (DateTimeFieldType orderedField: fieldOrder) {
			Tag tag = tagsToAddByType.get(orderedField);
			if (tag != null) {
				if (!newParts.isEmpty() && separator != null) {
					newParts.add(separator);
				}
				newParts.add(tag);
			}
		}
		return new DatePattern(newParts);
	}

	/**
	 * Derives a new date pattern by selecting only certain 
	 * desired tags, and the literals that are associated to them.
	 * @param fieldTypes fields to retain in new pattern
	 */
	public DatePattern withRetainedFields(EnumSet<DateTimeFieldType> fieldTypes) {
		List<PatternPart> resultParts = new ArrayList<PatternPart>();
		
		Literal currentLiteral = null;
		boolean empty = true;
		boolean first = true;

		// assumption: separators are associated with the component that comes after them
		// rule: literals between
		// eg:  d/M-w:y (remove w) => d/m:y   vs d/m-y
		for (PatternPart part: parts(true)) {
			// always add first literal
			if (part instanceof Literal) {
				if (first) {
					resultParts.add(part);
					empty = false;
					currentLiteral = null;
				}
				else {
					currentLiteral = (Literal) part;
				}
			}

			// add tags which haven't been excluded
			// (prepend with the literal that immediately preceds this tag,
			//  but only if something has been output already)
			else if (part instanceof Tag) {
				Tag tag = (Tag) part;
				if (fieldTypes.contains(tag.type())) {
					// add separators
					if (!empty && currentLiteral != null) {
						resultParts.add(currentLiteral);
					}
					
					// add components
					resultParts.add(tag);
					empty = false;
				}
				currentLiteral = null; // consumed whether used or not
			}
			
			// error
			else {
				throw new UnsupportedOperationException("Unknown pattern part "+part.getClass().getName());
			}
			
			first = false;
		}
		
		// always add if ends with literal
		if (currentLiteral != null) {
			resultParts.add(currentLiteral);
		}
		
		return new DatePattern(resultParts);
	}
	
	/**
	 * Parses the pattern.
	 * @param pattern the pattern to parse
	 * @return the compiled pattern 
	 * @throws NullPointerException if the given pattern is null
	 * @throws IllegalArgumentException if the given pattern is invalid
	 */
	private static List<PatternPart> parse(String pattern) {
		// sanity check: make sure it's valid
		new SimpleDateFormat(pattern);
		
		// parse into pattern parts
		List<PatternPart> parts = new ArrayList<PatternPart>();
		PeekableListIterator<Character> itr = PeekableListIterator.of(charListOf(pattern));
		while (itr.hasNext()) {
			// consume escaped quote character
			if (itr.hasNext(2) && itr.peekNext(1).charValue() == '\'' && itr.peekNext(2).charValue() == '\'') {
				itr.next(); // pop characters
				itr.next(); // pop characters
				parts.add(new Literal("'"));
			}

			// consume quoted literal
			else if (itr.peekNext().charValue() == '\'') {
				itr.next(); // pop character
				StringBuilder buf = new StringBuilder(); 
				while (itr.hasNext()) {
					// add escaped quote character
					if (itr.hasNext(2) && itr.peekNext(1).charValue() == '\'' && itr.peekNext(2).charValue() == '\'') {
						itr.next(); // pop characters
						itr.next(); // pop characters
						buf.append("'");
					}
					
					// end quoted literal
					else if (itr.peekNext().charValue() == '\'') {
						parts.add(new Literal(buf.toString()));
						itr.next(); // pop character
						break;
					}
					
					// continue literal
					else {
						buf.append(itr.next());
					}
				}
			}
			
			// consume un-quoted literal character
			else if (TAG_CHARS.indexOf(itr.peekNext()) < 0) {
				parts.add(new Literal(itr.next()));
			}
			
			// consume tag
			else {
				char ch = itr.next();
				StringBuilder buf = new StringBuilder();
				buf.append(ch);
				while (itr.hasNext() && itr.peekNext().charValue() == ch) {
					buf.append(itr.next());
				}
				parts.add(new Tag(buf.toString()));
			}
		}
		
		return parts;
	}
	
	/**
	 * Constructs a new list of parts where runs of consecutive literals
	 * are combined into one.
	 * The original list is not modified.
	 */
	private static List<PatternPart> contractLiteralsOf(List<PatternPart> parts) {
		List<PatternPart> result = new ArrayList<PatternPart>();
		List<Literal> currentRun = new ArrayList<Literal>();
		for (PatternPart part: parts) {
			if (part instanceof Literal) {
				// collect run of consecutive literals
				currentRun.add((Literal) part);
			}
			
			else {
				if (!currentRun.isEmpty()) {
					// add run
					if (currentRun.size() == 1) {
						result.add(currentRun.get(0));
					}
					else {
						result.add(new CompositeLiteral(currentRun));
					}
					currentRun.clear();
				}
				
				// add tag
				result.add(part);
			}
		}
		// add trailing run
		if (!currentRun.isEmpty()) {
			result.add(new CompositeLiteral(currentRun));
		}
		return result;
	}
	
	// convenience method turns a string into a list of characters,
	// for easy iteration over the characters
	private static List<Character> charListOf(String text) {
		List<Character> chars = new ArrayList<Character>();
		for (int i=0; i < text.length(); i++) {
			chars.add(text.charAt(i));
		}
		return chars;
	}
	
	/**
	 * Any component of the pattern.
	 */
	public static interface PatternPart {
		/**
		 * Gets the original pattern.
		 * Calls {@link #toPatternString()}.
		 */
		public String toString();
		
		/**
		 * Returns a string that indicates clearly what type of part,
		 * plus the value of the part.
		 */
		public String toSymbolicString();
	
		/**
		 * Returns a formatted string representation of this part,
		 * suitable for including straight into a full pattern string.
		 * Literal text is quoted where necessary.
		 */
		public String toPatternString();
		
		/**
		 * Gets the value of the part without any quoting or escaping.
		 * If the part is a literal quote character, then only
		 * that character will be returned.
		 */
		public String value();
	}
	
	/**
	 * Represents a single tag section within the pattern.
	 * Made up of a consecutive string of tag characters of the same type.
	 */
	public static class Tag implements PatternPart {
		private String value;
		private DateTimeFieldType fieldType;
		
		public Tag(String value) {
			this(DateTimeFieldType.valueOfCode(""+value.charAt(0)), value);
		}
	
		public Tag(DateTimeFieldType fieldType, String value) {
			this.value = value;
			this.fieldType = fieldType;
		}
		
		@Override
		public String toString() {
			return toPatternString();
		}
		
		@Override
		public String toSymbolicString() {
			return "TAG("+value+")";
		}
		
		@Override
		public String toPatternString() {
			return value;
		}
		
		/**
		 * Gets this tag code character.
		 * @return
		 */
		public char code() {
			return value.charAt(0);
		}
		
		public DateTimeFieldType type() {
			return fieldType;
		}
		
		/**
		 * Gets the full tag value, of a particular length.
		 * @return
		 */
		public String value() {
			return value;
		}
	
		public int length() {
			return value.length();
		}
		
		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result
					+ ((fieldType == null) ? 0 : fieldType.hashCode());
			result = prime * result + ((value == null) ? 0 : value.hashCode());
			return result;
		}
	
		@Override
		public boolean equals(Object obj) {
			if (this == obj) {
				return true;
			}
			if (obj == null) {
				return false;
			}
			if (!(obj instanceof Tag)) {
				return false;
			}
			Tag other = (Tag) obj;
			if (fieldType != other.fieldType) {
				return false;
			}
			if (value == null) {
				if (other.value != null) {
					return false;
				}
			} else if (!value.equals(other.value)) {
				return false;
			}
			return true;
		}
	}
	
	/**
	 * Represents a literal string within the pattern.
	 * The literal may or may not be quoted in the original pattern.
	 */
	public static class Literal implements PatternPart {
		protected String value;
		protected boolean needsQuote = false;
	
		public Literal(char ch) {
			this(""+ch);
		}
		
		public Literal(String value) {
			this.value = value;
	
			// all literal alphabetical characters are quoted
			for (int i=0; i < value.length(); i++) {
				char ch = value.charAt(i);
				if ((ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z')) {
					this.needsQuote = true;
				}
			}
		}
	
		@Override
		public String toString() {
			return toPatternString();
		}
		
		@Override
		public String toSymbolicString() {
			return "LITERAL("+value+")";
		}
		
		@Override
		public String toPatternString() {
			if (value.equals("'")) {
				return "''";
			}
			else {
				String text = value.replace("'", "''");
				if (needsQuote) {
					return "'"+text+"'";
				}
				else {
					return text;
				}
			}
		}
		
		public String value() {
			return value;
		}
	
		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + (needsQuote ? 1231 : 1237);
			result = prime * result + ((value == null) ? 0 : value.hashCode());
			return result;
		}
	
		@Override
		public boolean equals(Object obj) {
			if (this == obj) {
				return true;
			}
			if (obj == null) {
				return false;
			}
			if (!(obj instanceof Literal)) {
				return false;
			}
			Literal other = (Literal) obj;
			if (needsQuote != other.needsQuote) {
				return false;
			}
			if (value == null) {
				if (other.value != null) {
					return false;
				}
			} else if (!value.equals(other.value)) {
				return false;
			}
			return true;
		}
	}
	
	/**
	 * Represents a literal string within the pattern.
	 * The literal may or may not be quoted in the original pattern.
	 */
	public static class CompositeLiteral extends Literal {
		private List<Literal> contents;
	
		public CompositeLiteral(List<Literal> contents) {
			super(unquotedValueOf(contents));
			this.contents = new ArrayList<Literal>(contents);
		}
	
		@Override
		public String toString() {
			return toPatternString();
		}
		
		@Override
		public String toSymbolicString() {
			return super.toSymbolicString();
		}
		
		@Override
		public String toPatternString() {
			StringBuilder buf = new StringBuilder();
			for (Literal literal: contents) {
				buf.append(literal.toPatternString());
			}
			return buf.toString();
		}
		
		public String value() {
			return value;
		}
		
		private static String unquotedValueOf(List<Literal> literals) {
			StringBuilder buf = new StringBuilder();
			for (Literal literal: literals) {
				buf.append(literal.value());
			}
			return buf.toString();
		}
			
	}
}
