package net.sourceforge.tedhi;

/**
 * Defines a hint for how to default years when they are not supplied in the
 * input to a parser.
 * The year range bias is applied when neither date in a range are specified with a year.
 */
public enum YearRangeBias {
	/**
	 * Choose years such that the mid-point of the range is nearest the current date.
	 * This tends to be the most intuitive for date ranges which have no obvious
	 * inherent bias towards the past or future.
	 */
	NEAREST,
	
	/**
	 * Choose years such that at least one of the {@code from} and {@code to} dates
	 * are in the past or equal to the current day.
	 * The {@code from} year will be chosen such that the {@code from} date is the
	 * closest date to the current day that is either before or equal to the current day.
	 * The {@code to} date will be chosen as the same year or the following one, in order
	 * to maintain consistency.
	 */
	PAST,

	/**
	 * Choose years such that both the {@code from} and {@code to} dates are
	 * in the past or equal to the current day. 
	 * The {@code to} year will be chosen such that the {@code to} date is the
	 * closest date to the current day that is either before or equal to the current day.
	 * The {@code from} date will have the same year or the previous one, in order
	 * to maintain consistency.
	 */
	STRICTLY_PAST,
	
	/**
	 * Use the current year for both dates.
	 * Errors if this produces a range with the {@code from} date
	 * after the {@code to} date.
	 */
	CURRENT,
	
	/**
	 * Choose years such that at least one of the {@code from} and {@code to} dates
	 * are in the future or equal to the current day.
	 * The {@code to} year will be chosen such that the {@code to} date is the
	 * closest date to the current day that is either after or equal to the current day.
	 * The {@code from} date will have the same year or the following one, in order
	 * to maintain consistency.
	 */
	FUTURE,

	/**
	 * Choose years such that both the {@code from} and {@code to} dates are
	 * in the future or equal to the current day. 
	 * The {@code from} year will be chosen such that the {@code from} date is the
	 * closest date to the current day that is either after or equal to the current day.
	 * The {@code to} date will have the same year or the following one, in order
	 * to maintain consistency.
	 */
	STRICTLY_FUTURE,
	;
}
